#pragma once

// Copyright (C) 2014 Paul S. Heidmann
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// For a copy of the GNU General Public License see
// <http://www.gnu.org/licenses/>.
//
// Author contact info:
//   Paul Heidmann
//   paul@heidmann.com

#include <tuple>
#include <vector>
#include <cstdint>
#include <complex>
#include <functional>

namespace fractal
{

/// \brief This class is a container class for the parameters that are
///        used to generate a fractal.
///
/// This class is a container class for all of the parameters needed to
/// generate a fractal.
/// \tparam floatType The floating point type on which all parameters
///         are based.
/// \author Paul S. Heidmann
template<typename floatType>
class fractalParams
{
public:

  /// \brief This type is used to declare variables that hold the complex
  ///        numbers used in the fractal computations.
  typedef std::complex<floatType> cmplxType;

  /// \brief This type is used to declare variables that hold the function
  ///        that defines the factal (and its derivative).
  typedef std::function<cmplxType(const cmplxType&)> funcType;

  /// \brief This type is used to declare variables that hold the list
  ///        of zeros of the function being used to generate the
  ///        fractal (yes, these need to be known in advance).
  typedef std::vector<cmplxType> zerosLstType;

  /// \brief This type is used to declare variables that hold the color
  ///        of a given pixel in the fractal.
  typedef std::tuple<std::uint8_t, std::uint8_t, std::uint8_t> colorType;

  /// \brief This type is used to define variables that hold functions
  ///        that map fractal computation results to a color.
  ///
  /// In these functions, the first parameter is the maximum number
  /// of iterations allowed. The second parameter is the zero reached.
  /// The third parameter is the actual number of iterations.
  typedef std::function<colorType(
    const std::uint32_t,
    const std::complex<floatType>&,
    const std::uint32_t)> colorMapFuncType;

  /// \brief This constant attribute holds the lower left coordinate of the
  ///        fractal in the complex plain.
  const cmplxType lowerLeftExtent;

  /// \brief This constant attribute holds the upper right coordinate of the
  ///        fractal in the complex plain.
  const cmplxType upperRightExtent;

  /// \brief This constant attribute holds the distance between points in
  ///        the fractal. Note that this is a complex number, with the real
  ///        portion being the distance between the real components of two
  ///        neighboring points, and the imaginary component being the
  ///        distance between the imaginary components of two neighboring
  ///        points.
  const cmplxType delta;

  /// \brief This constant attribute holds the required accuracy that
  ///        Newton's method is required to produce when searching for
  ///        a zero.
  const floatType zeroEpsilon;

  /// \brief This constant attribute holds the function used to produce
  ///        the fractal.
  const funcType f;

  /// \brief This constant attribute holds the derivative of the function
  ///        used to produce the fractal.
  const funcType f_prime;

  /// \brief This constant attribute holds the list of zeros of the function
  ///        that is generating the fractal.
  const zerosLstType f_zeros;

  /// \brief This constant attribute holds the maximum number of iterations
  ///        that Newton's method is allowed.
  const std::uint32_t maxIterations;

  /// \brief This constant attribute holds the function that maps the output
  ///        of Newton's method to a color.
  const colorMapFuncType colorMapFunc;

  fractalParams() = delete;

  /// \brief This is the constructor that must be used to create instances
  ///        of this class.
  /// \note For a description of the parameters of this constructor, see
  ///       the descriptions of the public, constant attributes of this
  ///       class.
  fractalParams(
    const cmplxType& lwrLftExtnt,
    const cmplxType& upprRghtExtnt,
    const cmplxType& dlt,
    const floatType& zEpsln,
    const funcType& fractalFunc,
    const funcType& derivativeFractalFunc,
    const zerosLstType& fractalFuncZeros,
    const std::uint32_t maxIters,
    const colorMapFuncType& clrMapFunc ) :
      lowerLeftExtent{ lwrLftExtnt },
      upperRightExtent{ upprRghtExtnt },
      delta{ dlt },
      zeroEpsilon{ zEpsln },
      f{ fractalFunc },
      f_prime{ derivativeFractalFunc },
      f_zeros{ fractalFuncZeros },
      maxIterations{ maxIters },
      colorMapFunc{ clrMapFunc }
  {
  }

  /// \brief The class destructor.
  ~fractalParams() {}

};

}
