// Copyright (C) 2014 Paul S. Heidmann
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// For a copy of the GNU General Public License see
// <http://www.gnu.org/licenses/>.
//
// Author contact info:
//   Paul Heidmann
//   paul@heidmann.com

#include <fractalTwo.hpp>

#include <stdexcept>

namespace FT=fractalTwo;

namespace
{
  constexpr FT::fltType pi{ 3.1415926535 };
  constexpr FT::fltType twoPi{ 2.0 * pi };
  constexpr FT::fltType threePi{ 3.0 * pi };
  constexpr FT::fltType fourPi{ 4.0 * pi };
  constexpr FT::fltType piOverTwo{ pi / 2.0 };

  FT::fractalParamsType::colorType colorForZero(
    const std::uint32_t maxIters,
    const FT::cmplxType& z,
    const std::uint32_t numIters )
  {
    const FT::fltType scaleFactor{
      0.75 * (static_cast<FT::fltType>(numIters) / static_cast<FT::fltType>(maxIters)) + 0.25 };
    const FT::fltType colorMagnitude{
      static_cast<FT::fltType>(0xff) * scaleFactor };
    const std::uint32_t color{ static_cast<std::uint32_t>(colorMagnitude) };

    if( std::abs( z + pi ) < 0.1 )
      return( std::move( std::make_tuple(
			   color, 0x00, color ) ) );
    if( std::abs( z ) < 0.1 )
      return( std::move( std::make_tuple(
			   color, color, 0x00 ) ) );
    if( std::abs( z - twoPi ) < 0.1 )
      return( std::move( std::make_tuple(
			   color, 0x00, 0x00 ) ) );
    if( std::abs( z - threePi ) < 0.1 )
      return( std::move( std::make_tuple(
			   0x00, color, 0x00 ) ) );
    if( std::abs( z - fourPi ) < 0.1 )
      return( std::move( std::make_tuple(
			   0x00, 0x00, color ) ) );
    throw( std::runtime_error( "Unknown zero..." ) );
  }

  FT::cmplxType f( const FT::cmplxType& z )
  {
    // Use zeros at -pi, 0, pi, 2*pi, 3*pi, 4*pi
    return( std::sin<FT::fltType>( z ) );
  }

  FT::cmplxType f_prime( const FT::cmplxType& z )
  {
    return( std::cos<FT::fltType>( z ) );
  }
}

std::shared_ptr<FT::fractalParamsType> FT::getFractalParams( void )
{
#if 0
  const FT::cmplxType lowerLeft{ (piOverTwo - 0.3), (piOverTwo - 0.65) };
  const FT::cmplxType upperRight{ (piOverTwo + 0.3), (piOverTwo - 0.05) };
  const FT::cmplxType delta{ 0.0001, 0.0001 };
#else
  const FT::cmplxType lowerLeft{ (piOverTwo - 0.06), (piOverTwo - 0.47) };
  const FT::cmplxType upperRight{ (piOverTwo + 0.06), (piOverTwo - 0.23) };
  const FT::cmplxType delta{ 0.00003, 0.00003 };
#endif

  const FT::fltType zEpsilon{ 0.05 };
  const std::vector<FT::cmplxType> zeros{
    FT::cmplxType{ -pi, 0.0 },
    FT::cmplxType{ 0.0, 0.0 },
    FT::cmplxType{ twoPi, 0.0 },
    FT::cmplxType{ threePi, 0.0 },
    FT::cmplxType{ fourPi, 0.0 } };
  const std::uint32_t maxIters{ 50 };

  const std::shared_ptr<FT::fractalParamsType> fParams{
    std::make_shared<FT::fractalParamsType>(
      lowerLeft,
      upperRight,
      delta,
      zEpsilon,
      f,
      f_prime,
      zeros,
      maxIters,
      colorForZero ) };
  return( fParams );
}
