#pragma once

// Copyright (C) 2014 Paul S. Heidmann
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// For a copy of the GNU General Public License see
// <http://www.gnu.org/licenses/>.
//
// Author contact info:
//   Paul Heidmann
//   paul@heidmann.com

#include <complex>
#include <cstdint>
#include <stdexcept>
#include <functional>

namespace fractal
{

/// \brief This class implements Newton's method (the same one from
///        your typical first year calculus class), to find zeros
///        of complex valued functions (in the complex plain,
///        obviously).
/// \tparam floatType The floating point type to be used in the
///         underlying computations.
/// \author Paul S. Heidmann
template<typename floatType>
class newtonsMethod
{
public:

  /// \brief This type is used to declare variables that hold
  ///        complex numbers.
  typedef std::complex<floatType> cmplxType;

  /// \brief This type is used to declare variables that hold
  ///        functions whose zeros are being found.
  typedef std::function<cmplxType(const cmplxType&)> funcType;

  newtonsMethod() = delete;

  /// \brief This is the constructor that must be used to create
  ///        instances of this class.
  /// \param initialGuess The complex number that is to serve as
  ///        the initial guess of the zero to be found.
  /// \param f The function whose zero is to be found.
  /// \param f_prime The derivative of the function whose zero is
  ///        to be found.
  /// \param maxIterations The maximum number of iterations that may
  ///        be used by this implementation to find a zero.
  /// \param epsilon The required accuracy. The zero must be found
  ///        within this tolerance.
  newtonsMethod(
    const cmplxType& initialGuess,
    const funcType& f,
    const funcType& f_prime,
    const std::uint32_t maxIterations,
    const floatType& epsilon ) :
      m_initialGuess{ initialGuess },
      m_f{ f },
      m_fPrime{ f_prime },
      m_maxIterations{ maxIterations },
      m_epsilon{ epsilon }
  {
  }

  /// \brief The class destructor.
  ~newtonsMethod() {}

  /// \brief This method is called to find a zero of the given function,
  ///        using the given initial guess.
  /// \param numIters An output parameter. The number of iterations actually
  ///        required to find the given zero.
  /// \return The zero found.
  const cmplxType findZero( std::uint32_t& numIters ) const
  {
    numIters = 0;
    if( (this->m_epsilon) < 0.0 )
      throw std::runtime_error( "Epsilon must be greater than zero!" );

    cmplxType curGuess( this->m_initialGuess );
    while( numIters < (this->m_maxIterations) )
    {
      const cmplxType prevGuess{ curGuess };
      const cmplxType numerator{ (this->m_f)(curGuess) };
      const cmplxType denominator{ (this->m_fPrime)(curGuess) };
      if( denominator == 0.0 )
	throw std::runtime_error( "Zero denominator in newtons method!" );
      curGuess -= (numerator / denominator);
      const floatType curEpsilon{ std::abs( curGuess - prevGuess ) };
      if( curEpsilon <= (this->m_epsilon) )
	return( std::move( curGuess ) );
      ++numIters;
    }
    throw std::runtime_error( "Newton's method not converging!" );
  }

private:

  /// \brief This attribute holds the caller supplied initial guess.
  const cmplxType m_initialGuess;

  /// \brief This attribute holds the function whose zero is to be found.
  const funcType m_f;

  /// \brief This attribute holds the derivative of the function whose
  ///        zero is to be found.
  const funcType m_fPrime;

  /// \brief This attribute holds the maximum number of iterations allowed
  ///        when searching for a zero.
  const std::uint32_t m_maxIterations;

  /// \brief This attribute holds the required accuracy (that is, tolerance).
  const floatType m_epsilon;
};

}
