// Copyright (C) 2014 Paul S. Heidmann
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// For a copy of the GNU General Public License see
// <http://www.gnu.org/licenses/>.
//
// Author contact info:
//   Paul Heidmann
//   paul@heidmann.com

#include "pngFactory.hpp"

#include <algorithm>
#include <functional>
#include <stdio.h>

png::pngFactory::pngFactory(
  const std::uint16_t xResolution,
  const std::uint16_t yResolution,
  const std::string& pngOutFileName ) :
  pFileName( pngOutFileName ),
  xRes( xResolution ),
  yRes( yResolution ),
  xMin( 1.0e300 ),
  xMax( -1.0e300 ),
  yMin( 1.0e300 ),
  yMax( -1.0e300 ),
  xTranslationFactor( 0.0 ),
  yTranslationFactor( 0.0 ),
  pngRowPointers( 0 )
{
  this->pxlLst.clear();
  this->allocatePngImageSpace();
}

png::pngFactory::~pngFactory()
{
  this->findMinMax();
  this->translateAllPixels();
  this->renderAllPixels();
  this->writePngFile();
  this->deallocatePngImageSpace();
}

void png::pngFactory::addPixel(
  const long double x,
  const long double y,
  const std::uint8_t red,
  const std::uint8_t green,
  const std::uint8_t blue )

{
  this->pxlLst.push_back( std::make_tuple(
			    x,
			    y,
			    red,
			    green,
			    blue ) );
}

const std::uint16_t png::pngFactory::getXRes( void ) const
{
  return( this->xRes );
}

const std::uint16_t png::pngFactory::getYRes( void ) const
{
  return( this->yRes );
}

void png::pngFactory::allocatePngImageSpace( void )
{
  this->pngRowPointers =
    new png::pngFactory::pngRowPtrType[this->yRes];
  for( std::uint16_t i = 0; i < this->yRes; ++i )
    this->pngRowPointers[i] =
      new std::uint8_t[(3 * (this->xRes + 1))];
}

void png::pngFactory::deallocatePngImageSpace( void )
{
  for( std::uint16_t i = 0; i < this->yRes; ++i )
    delete[]( this->pngRowPointers[i] );
  delete[]( this->pngRowPointers );
}

void png::pngFactory::writePngFile( void )
{
  ::FILE* pngOutFile( fopen( this->pFileName.c_str(), "wb" ) );

  this->pngStructPtr = png_create_write_struct(
    PNG_LIBPNG_VER_STRING, NULL, NULL, NULL );

  this->pngInfoPtr = png_create_info_struct( this->pngStructPtr );

  png_init_io( this->pngStructPtr, pngOutFile );

  if ( setjmp( png_jmpbuf( this->pngStructPtr ) ) )
  {
  }

  png_set_IHDR( this->pngStructPtr,
		this->pngInfoPtr,
		this->xRes,
		this->yRes,
		8,
		PNG_COLOR_TYPE_RGB,
		PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_BASE,
		PNG_FILTER_TYPE_BASE );

  png_write_info( this->pngStructPtr, this->pngInfoPtr );

  if ( setjmp( png_jmpbuf( this->pngStructPtr ) ) )
  {
  }

  png_write_image( this->pngStructPtr,
		   reinterpret_cast<png_byte**>( this->pngRowPointers ) );

  if( setjmp(png_jmpbuf( this->pngStructPtr ) ) )
  {
  }

  png_write_end( this->pngStructPtr, NULL );

  fclose( pngOutFile );
}

void png::pngFactory::updateMinMax(
  const png::pngFactory::pixelType& p )
{
  const long double curX( std::get<0>(p) );
  const long double curY( std::get<1>(p) );
  if( this->xMin > curX )
    this->xMin = curX;
  if( this->xMax < curX )
    this->xMax = curX;
  if( this->yMin > curY )
    this->yMin = curY;
  if( this->yMax < curY )
    this->yMax = curY;
}

void png::pngFactory::findMinMax( void )
{
  std::for_each(
    this->pxlLst.begin(),
    this->pxlLst.end(),
    std::bind(
      &png::pngFactory::updateMinMax,
      this,
      std::placeholders::_1 ) );
}

void png::pngFactory::translatePixel(
  png::pngFactory::pixelType& p ) const
{
  std::get<0>(p) = this->xTranslationFactor * (std::get<0>(p) - this->xMin);
  std::get<1>(p) = this->yTranslationFactor * (std::get<1>(p) - this->yMin);
}

void png::pngFactory::translateAllPixels( void )
{
  this->xTranslationFactor =
    static_cast<long double>( this->xRes ) / (this->xMax - this->xMin);
  this->yTranslationFactor =
    static_cast<long double>( this->yRes ) / (this->yMax - this->yMin);
  std::for_each(
    this->pxlLst.begin(),
    this->pxlLst.end(),
    std::bind(
      &png::pngFactory::translatePixel,
      this,
      std::placeholders::_1 ) );
}

void png::pngFactory::renderAllPixels( void )
{
  const std::list<png::pngFactory::pixelType>::const_iterator
    end( this->pxlLst.end() );

  std::list<png::pngFactory::pixelType>::const_iterator i;
  for( i = this->pxlLst.begin();
       i != end;
       ++i )
  {
    const int x( static_cast<int>( std::get<0>(*i) ) );
    const int y( static_cast<int>( std::get<1>(*i) ) );

    if( (x < this->xRes) && (y < this->yRes) &&
	(x >= 0) && (y >= 0) )
    {
      std::uint8_t*
	pix( &(this->pngRowPointers[y][3*x]) );
      *pix = std::get<2>(*i);
      ++pix;
      *pix = std::get<3>(*i);
      ++pix;
      *pix = std::get<4>(*i);
      ++pix;
    }
  }
}
