#pragma once

// Copyright (C) 2014 Paul S. Heidmann
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// For a copy of the GNU General Public License see
// <http://www.gnu.org/licenses/>.
//
// Author contact info:
//   Paul Heidmann
//   paul@heidmann.com

#include <list>
#include <tuple>
#include <string>
#include <cstdint>
#include <boost/utility.hpp>
#include <boost/tuple/tuple.hpp>

#include <png.h>

namespace png
{

/// \brief This class provides an interface into libpng that is not
///        completely insane (like libpng's native interface). The
///        code in this class is fairly ugly.
class pngFactory : private boost::noncopyable
{
public:

  pngFactory() = delete;

  /// \brief This is the constructor that must be used to create
  ///        instances of this class.
  /// \param xResolution The horizontal resolution of the image.
  /// \param yResolution The vertical resolution of the image.
  /// \param pngOutFileName The filename of the final PNG image.
  pngFactory( const std::uint16_t xResolution,
	      const std::uint16_t yResolution,
	      const std::string& pngOutFileName );

  /// \brief The class destructor.
  ~pngFactory();

  /// \brief This method is called to add a pixel to the PNG image.
  /// \param x The X coordinate of the pixel.
  /// \param y The Y coordinate of the pixel.
  /// \param red The red component of the color of the pixel.
  /// \param green The green component of the color of the pixel.
  /// \param blue The blue component of the color of the pixel.
  void addPixel(
    const long double x,
    const long double y,
    const std::uint8_t red,
    const std::uint8_t green,
    const std::uint8_t blue );

  /// \brief This method returns to the caller the horizontal resolution
  ///        of the image.
  /// \return The X resolution of the image.
  const std::uint16_t getXRes( void ) const;

  /// \brief This method returns to the caller the vertical resolution
  ///        of the image.
  /// \return The Y resolution of the image.
  const std::uint16_t getYRes( void ) const;

private:

  /// \brief This type is used to define variables that hold single
  ///        pixels.
  typedef std::tuple<
    long double,
    long double,
    std::uint8_t,
    std::uint8_t,
    std::uint8_t> pixelType;

  /// \brief This type is used to access rows in the PNG image.
  typedef std::uint8_t* pngRowPtrType;

  /// \brief This attribute holds the name of the file into which the
  ///        final PNG image is to placed.
  const std::string pFileName;

  /// \brief This attribute holds the X resolution of the PNG image.
  const std::uint16_t xRes;

  /// \brief This attribute holds the Y resolution of the PNG image.
  const std::uint16_t yRes;

  /// \brief This attribute holds the smallest x coordinate the caller
  ///        has specified.
  long double xMin;

  /// \brief This attribute holds the largest x coordinate the caller
  ///        has specified.
  long double xMax;

  /// \brief This attribute holds the smallest y coordinate the caller
  ///        has specified.
  long double yMin;

  /// \brief This attribute holds the largest y coordinate the caller
  ///        has specified.
  long double yMax;

  /// \brief This attribute holds the scalar used to translate caller
  ///        specified x coordinates into the image coordinate space.
  long double xTranslationFactor;

  /// \brief This attribute holds the scalar used to translate caller
  ///        specified y coordinates into the image coordinate space.
  long double yTranslationFactor;

  /// \brief This attribute holds all caller specified pixels.
  std::list<png::pngFactory::pixelType> pxlLst;

  /// \brief This attribute holds a row pointer used to access the
  ///        PNG image.
  png::pngFactory::pngRowPtrType* pngRowPointers;

  /// \brief This attribute holds the libpng defined PNG structure used
  ///        when generating the PNG image.
  ::png_structp pngStructPtr;

  /// \brief This attribute holds the libpng defined PNG info structure used
  ///        when generating the PNG image.
  ::png_infop pngInfoPtr;

  /// \brief This method allocates space from the heap to hold the PNG
  ///        image.
  void allocatePngImageSpace( void );

  /// \brief This method deallocates the PNG image space.
  void deallocatePngImageSpace( void );

  /// \brief This method writes the PNG image out to disk.
  void writePngFile( void );

  /// \brief This method is called to update the minimums and maximums,
  ///        as tracked by this instance.
  void updateMinMax( const png::pngFactory::pixelType& p );

  /// \brief This method is called to determine the minimums and maximums,
  ///        as tracked by this instance.
  void findMinMax( void );

  /// \brief This method maps a single pixel from the user specified
  ///        coordinate plane to pixel space.
  /// \param p The pixel to translate.
  void translatePixel( png::pngFactory::pixelType& p ) const;

  /// \brief This method maps all pixels from the user specified
  ///        coordinate plane to pixel space.
  void translateAllPixels( void );

  /// \brief This method renders all pixels.
  void renderAllPixels( void );
};

};
