#pragma once

// Copyright (C) 2014 Paul S. Heidmann
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// For a copy of the GNU General Public License see
// <http://www.gnu.org/licenses/>.
//
// Author contact info:
//   Paul Heidmann
//   paul@heidmann.com

#include <deque>
#include <mutex>
#include <vector>
#include <memory>
#include <thread>
#include <bitset>
#include <cstdint>
#include <iostream>
#include <algorithm>
#include <functional>
#include <boost/utility.hpp>

#include <math.h>  // for fabs

#include <pngFactory.hpp>
#include <fractalParams.hpp>
#include <newtonsMethod.hpp>

namespace fractal
{

/// \brief This class splits the work of computing a given fractal amoung
///        several threads.
///
/// This class splits the work of computing a given fractal among several
/// threads. This is done to take advantage of machines with multiple cores.
/// The number of threads spawned is settable (see m_numThreads). The
/// fractal is split by rows (that is, lines with a constant imaginary
/// component in the complex plane), with each row being the unit of
/// work for a thread. Computation continues by all threads until all
/// rows are computed.
/// \tparam floatType The underlying floating point type to be used in
///                   all computations. This will also be the type used
///                   to instantiate all complex numbers.
/// \author Paul S. Heidmann
template <typename floatType>
class threadedFractalEngine : private boost::noncopyable
{
public:

  /// \brief This type is used to define variables that hold colors.
  ///        The standard RGB color definition is used here.
  typedef std::tuple<std::uint8_t, std::uint8_t, std::uint8_t> colorType;

  threadedFractalEngine() = delete;

  /// \brief The class destructor.
  ~threadedFractalEngine() {}

  /// \brief This is the constructor that must be used to create instances
  ///        of this class.
  /// \param fParams The parameters used to generate the given fractal.
  /// \param pngFact The PNG factory class used to create the PNG image.
  threadedFractalEngine(
    const std::shared_ptr<fractal::fractalParams<floatType>>& fParams,
    const std::shared_ptr<png::pngFactory>& pngFact ) :
      m_numThreads{ 8 },
      m_fParams{ fParams },
      m_pngFact{ pngFact },
      m_tenthOfRange{ 0 }
  {
  }

  /// \brief This method is called to compute the given fractal.
  void doFractal( void )
  {
    // First, load up the row queue, before spawning any worker threads.
    // That way, the queue being empty will always indicate that all
    // work has been completed.
    {
      std::lock_guard<std::mutex> lk{ this->m_rowQueueGuard };
      for( floatType im{ this->m_fParams->lowerLeftExtent.imag() };
	   im < this->m_fParams->upperRightExtent.imag();
	   im += this->m_fParams->delta.imag() )
	this->m_rowQueue.push_back( im );
    }

    // Compute tenths of the range. This will be for the status count.
    this->m_tenthOfRange = (fabs(
      this->m_fParams->upperRightExtent.imag() -
      this->m_fParams->lowerLeftExtent.imag() )) / 10.0;
    std::cout << "Count to ten:" << std::endl;

    // Spawn the worker threads.
    for( std::uint32_t thr{0}; thr < (this->m_numThreads); ++thr )
    {
      std::unique_ptr<std::thread> curThr{
	new std::thread{
	  std::bind(
	    &threadedFractalEngine<floatType>::fractalWorkerThread,
	    this ) } };
      this->m_threads.push_back( std::move(curThr) );
    }

    // Wait for each thread to finish. When all threads have
    // exitted, the work is complete.
    std::for_each(
      this->m_threads.begin(),
      this->m_threads.end(),
      []( decltype(*(this->m_threads.begin()))& thrd ) -> void
    {
      thrd->join();
    } );
    std::cout << std::endl;
  }

private:

  /// \brief This attribute holds the number of threads that this instance
  ///        is to spawn to compute the given fractal.
  const std::uint32_t m_numThreads;

  /// \brief This attribute holds a pointer to the class instance that
  ///        contains the parameters used to generate the fractal.
  const std::shared_ptr<fractal::fractalParams<floatType>> m_fParams;

  /// \brief This attribute holds a pointer to the factory used to generate
  ///        the PNG image.
  const std::shared_ptr<png::pngFactory> m_pngFact;

  /// \brief This attribute holds a mutex used to serialize access to the
  ///        PNG factory.
  std::mutex m_pngFactGuard;

  /// \brief This attribute is the queue that holds the rows of the fractal
  ///        yet to be computed.
  std::deque<floatType> m_rowQueue;

  /// \brief This attribute holds the mutex that guards access to the
  ///        row queue.
  /// \sa m_rowQueue
  std::mutex m_rowQueueGuard;

  /// \brief This attribute holds the thread objects used to compute the
  ///        given fractal.
  std::vector<std::unique_ptr<std::thread>> m_threads;

  /// \brief This attribute holds a tenth of the imaginary range (that is,
  ///        a tenth of the rows). This is used in the status count.
  floatType m_tenthOfRange;

  /// \brief This attribute holds ten booleans. These booleans are set when
  ///        the corresponding status number has been printed. Note that the
  ///        default constructor sets all bits to false, and this class takes
  ///        advantage of this fact.
  std::bitset<10> m_tenthStatusPrinted;

  /// \brief This method is the main (top level) method of each of the fractal
  ///        computation threads.
  void fractalWorkerThread( void )
  {
    floatType curRow{ 0.0 };
    std::int32_t statNum{ -1 };
    while( true )
    {
      {
	std::lock_guard<std::mutex> lk{ this->m_rowQueueGuard };
	if( this->m_rowQueue.empty() )
	  return;
	curRow = this->m_rowQueue.front();
	this->m_rowQueue.pop_front();

	// See if it is time to print a status number. Note that this is
	// done while the mutex is still locked, for obvious reasons.
	const std::uint32_t tenth(
	  static_cast<std::uint32_t>(
	    (curRow - this->m_fParams->lowerLeftExtent.imag()) / (this->m_tenthOfRange) ) );
	if( !(this->m_tenthStatusPrinted[tenth]) )
	{
	  this->m_tenthStatusPrinted.set( tenth, true );
	  statNum = tenth;
	}
	else
	{
	  statNum = -1;
	}
      }

      if( statNum >= 0 )
	std::cout << statNum << std::flush;

      for( floatType rl{ this->m_fParams->lowerLeftExtent.real() };
	   rl < this->m_fParams->upperRightExtent.real();
	   rl += this->m_fParams->delta.real() )
      {
	try
	{
	  const fractal::newtonsMethod<floatType> solver{
	    std::complex<floatType>( rl, curRow ),
	    this->m_fParams->f,
	    this->m_fParams->f_prime,
	    this->m_fParams->maxIterations,
	    this->m_fParams->zeroEpsilon };
	  std::uint32_t numIterations{ 0 };
	  const std::complex<floatType> zero{
	    solver.findZero( numIterations ) };
	  const colorType color{
	    this->m_fParams->colorMapFunc(
	      this->m_fParams->maxIterations,
	      zero,
	      numIterations ) };
	  {
	    std::lock_guard<std::mutex> lk{ this->m_pngFactGuard };
	    this->m_pngFact->addPixel(
	      rl,
	      curRow,
	      std::get<0>(color),
	      std::get<1>(color),
	      std::get<2>(color) );
	  }
	}
	catch( ... )
	{
	  std::lock_guard<std::mutex> lk{ this->m_pngFactGuard };
	  this->m_pngFact->addPixel(
	    rl,
	    curRow,
	    0,
	    0,
	    0 );
	}
      }
    }
  }
};

}
